theory Typing
imports Syntax

begin

consts qualRel :: "qual \<Rightarrow> qual \<Rightarrow> bool" (infix "\<preceq>" 50)
primrec
  "qU \<preceq> q = True"
  "qL \<preceq> q = (q = qL)"

consts qualT :: "ty \<Rightarrow> qual"
primrec
  "qualT (Qual q \<phi>) = q"

datatype Term = Term "ty"
types env = "(name * Term) set"

syntax
  var_term :: "name \<Rightarrow> Term \<Rightarrow> (name * Term)"  ("(_\<rhd>_)" [90,90] 90)
translations
  "(x\<rhd>\<tau>)" \<rightleftharpoons> "(x, Term \<tau>)"

syntax
  addbinding :: "(name * Term) set \<Rightarrow> name \<Rightarrow> ty \<Rightarrow> (name * Term) set"
    ("_','(_':_')" [90,0,0] 91)
translations
  "\<Gamma>,(x:\<tau>)" \<rightleftharpoons> "insert (x\<rhd>\<tau>) \<Gamma>"

constdefs un :: "env \<Rightarrow> env"
 "un \<Gamma> \<equiv> { (x\<rhd>\<tau>) | x \<tau>. (x\<rhd>\<tau>) \<in> \<Gamma> \<and> qualT \<tau> \<preceq> qU }"

constdefs lin :: "env \<Rightarrow> env"
 "lin \<Gamma> \<equiv> { (x\<rhd>\<tau>) | x \<tau>. (x\<rhd>\<tau>) \<in> \<Gamma> \<and> qL \<preceq> qualT \<tau> }"

constdefs qualGrelL :: "env \<Rightarrow> qual \<Rightarrow> bool" ("qualG _ \<preceq> _" [90,90] 50)
  "qualGrelL \<Gamma> q \<equiv> (\<forall>x \<tau>. (x\<rhd>\<tau>) \<in> \<Gamma> \<longrightarrow> qualT \<tau> \<preceq> q)"

constdefs qualGrelR :: "qual \<Rightarrow> env \<Rightarrow> bool" ("_ \<preceq> qualG _" [90,90] 50)
  "qualGrelR q \<Gamma> \<equiv> (\<forall>x \<tau>. (x\<rhd>\<tau>) \<in> \<Gamma> \<longrightarrow> q \<preceq> qualT \<tau>)"

lemma qualT_unrestricted:
  assumes "qualT \<tau> \<preceq> qU"
  shows "qualT \<tau> = qU"
using prems by (cases "qualT \<tau>", auto)

lemma qualG_relL_subset:
  assumes "qualG \<Gamma> \<preceq> q"
  and "\<Gamma>' \<subseteq> \<Gamma>"
  shows "qualG \<Gamma>' \<preceq> q"
using prems by (auto simp add: qualGrelL_def)

lemma equal_unrestricted:
  assumes "qualG \<Gamma> \<preceq> qU"
  shows "un \<Gamma> = \<Gamma>"
using prems proof auto
  fix x b
  assume "(x, b) \<in> \<Gamma>"
  then obtain \<tau> where "b = Term \<tau>" by (cases b)
  hence "(x\<rhd>\<tau>) \<in> \<Gamma>" and "qualT \<tau> \<preceq> qU" using prems by (auto simp add: qualGrelL_def)
  thus "(x, b) \<in> un \<Gamma>" using prems by (auto simp add: un_def)
qed (auto simp add: un_def)

constdefs wf_env :: "env \<Rightarrow> bool" ("_ \<Turnstile> OK" [50] 50)
  "wf_env \<Gamma> \<equiv> finite \<Gamma> \<and> (\<forall>x \<tau>1 \<tau>2. (x\<rhd>\<tau>1) \<in> \<Gamma> \<and> (x\<rhd>\<tau>2) \<in> \<Gamma> \<longrightarrow> \<tau>1 = \<tau>2)"

lemma wf_env_subset:
  assumes "\<Gamma> \<Turnstile> OK"
  and "\<Gamma>' \<subseteq> \<Gamma>"
  shows "\<Gamma>' \<Turnstile> OK"
using prems finite_subset[of \<Gamma>' \<Gamma>] by (unfold wf_env_def, fast)

lemma env_cons_without_eq_left:
  assumes "\<Gamma>1 \<union> \<Gamma>2 \<Turnstile> OK"
  and "(y\<rhd>\<tau>) \<in> \<Gamma>2"
  shows "\<Gamma>1,(y:\<tau>) \<union> \<Gamma>2\<bar>y = \<Gamma>1 \<union> \<Gamma>2"
using prems proof (auto simp add: wf_env_def without_def)
  fix x b
  assume "(x\<rhd>\<tau>) \<in> \<Gamma>2" and "(x, b) \<in> \<Gamma>2" and "b \<noteq> Term \<tau>"
  have "\<Gamma>2 \<Turnstile> OK" using prems wf_env_subset[of "\<Gamma>1 \<union> \<Gamma>2" \<Gamma>2] by auto
  hence "b = Term \<tau>" using prems by (cases b, auto simp add: wf_env_def)
  thus False by contradiction
qed

lemma env_cons_without_eq_right:
  assumes "\<Gamma>1 \<union> \<Gamma>2 \<Turnstile> OK"
  and "(y\<rhd>\<tau>) \<in> \<Gamma>1"
  shows "\<Gamma>1\<bar>y \<union> \<Gamma>2,(y:\<tau>) = \<Gamma>1 \<union> \<Gamma>2"
proof -
  have "\<Gamma>2 \<union> \<Gamma>1 \<Turnstile> OK" using prems by (auto simp add: Un_commute)
  hence "\<Gamma>2,(y:\<tau>) \<union> \<Gamma>1\<bar>y = \<Gamma>2 \<union> \<Gamma>1" using prems env_cons_without_eq_left by auto
  thus ?thesis by (auto simp add: Un_commute)
qed

lemma wf_env_cons:
  assumes "\<Gamma> \<Turnstile> OK"
  and "\<not> \<Gamma> defines x"
  shows "\<Gamma>,(x:\<tau>) \<Turnstile> OK"
using prems by (auto simp add: dom_def wf_env_def)

lemma wf_env_union_sanity:
  assumes "\<Gamma>1\<union>\<Gamma>2 \<Turnstile> OK"
  and "(x\<rhd>\<tau>) \<in> \<Gamma>2"
  and "\<Gamma>1 defines x"
  shows "(x\<rhd>\<tau>) \<in> \<Gamma>1"
proof -
  obtain b where "(x, b) \<in> \<Gamma>1" using prems by (auto simp add: dom_def)
  then obtain \<tau>' where "(x\<rhd>\<tau>') \<in> \<Gamma>1" by (cases b, auto)
  also hence "\<tau>' = \<tau>" using prems by (auto simp add: wf_env_def)
  finally show ?thesis .
qed

lemma wf_env_union_sanity_cons:
  assumes "\<Gamma>1\<union>\<Gamma>2 \<Turnstile> OK"
  and "(x\<rhd>\<tau>) \<in> \<Gamma>1"
  shows "\<Gamma>2,(x:\<tau>) \<Turnstile> OK"
using prems proof (cases "\<Gamma>2 defines x")
  case False
  have "\<Gamma>2 \<Turnstile> OK" using prems wf_env_subset[of "\<Gamma>1\<union>\<Gamma>2" \<Gamma>2] by auto
  thus ?thesis using False wf_env_cons by auto
next
  case True
  then obtain b where "(x, b) \<in> \<Gamma>2" using prems by (auto simp add: dom_def)
  then obtain \<tau>' where "(x\<rhd>\<tau>') \<in> \<Gamma>2" by (cases b, auto)
  also hence "\<tau>' = \<tau>" using prems by (auto simp add: wf_env_def)
  finally have "\<Gamma>2,(x:\<tau>) = \<Gamma>2" by auto
  thus ?thesis using prems wf_env_subset[of "\<Gamma>1\<union>\<Gamma>2" \<Gamma>2] by auto
qed

lemma wf_env_subset_defines:
  assumes Hok: "\<Gamma>s \<Turnstile> OK"
  and Hsub: "\<Gamma>r \<subseteq> \<Gamma>s"
  and Hx: "(x\<rhd>\<tau>) \<in> \<Gamma>s"
  and Hd: "\<Gamma>r defines x"
  shows "(x\<rhd>\<tau>) \<in> \<Gamma>r"
proof -
  obtain b where "(x, b) \<in> \<Gamma>r" using Hd defines_sanity by auto
  then obtain \<tau>' where "(x\<rhd>\<tau>') \<in> \<Gamma>r" by (cases b, auto)
  hence "(x\<rhd>\<tau>') \<in> \<Gamma>s" using Hsub by auto
  also hence "\<tau>' = \<tau>" using Hx Hok by (auto simp add: wf_env_def)
  thus ?thesis using prems by auto
qed

lemma un_without_l_sanity:
  assumes "\<Gamma> \<Turnstile> OK"
  and "(x\<rhd>\<tau>) \<in> \<Gamma>"
  and "qualT \<tau> = qL"
  shows "un \<Gamma> = un (\<Gamma>\<bar>x)"
proof (auto simp add: un_def without_def)
  fix \<tau>'
  assume "(x\<rhd>\<tau>') \<in> \<Gamma>" and "qualT \<tau>' \<preceq> qU"
  hence "\<tau> = \<tau>'" using prems by (auto simp add: wf_env_def)
  hence "qualT \<tau> = qualT \<tau>'" by auto
  hence "qL \<preceq> qU" using prems by auto
  thus False by auto
qed

constdefs env_compat :: "env \<Rightarrow> env \<Rightarrow> bool"  (infix "\<sim>" 50)
   "\<Gamma>1 \<sim> \<Gamma>2 \<equiv> \<forall>x \<tau>. (x\<rhd>\<tau>) \<in> (\<Gamma>1 \<inter> \<Gamma>2) \<longrightarrow> qualT \<tau> \<preceq> qU"

lemma env_compat_weakening_left:
  assumes "\<Gamma>1 \<sim> \<Gamma>2"
  and "qualG \<Gamma> \<preceq> qU"
  shows "\<Gamma>1 \<union> \<Gamma> \<sim> \<Gamma>2"
using prems by (auto simp add: env_compat_def qualGrelL_def)

lemma env_compat_symm:
  assumes "\<Gamma>1 \<sim> \<Gamma>2"
  shows "\<Gamma>2 \<sim> \<Gamma>1"
using prems by (auto simp add: env_compat_def)

lemma env_compat_union_left:
  assumes "\<Gamma> \<sim> \<Gamma>1"
  and "\<Gamma> \<sim> \<Gamma>2"
  shows "\<Gamma> \<sim> \<Gamma>1 \<union> \<Gamma>2"
using prems by (auto simp add: env_compat_def)

lemma env_compat_union_right:
  assumes "\<Gamma>1 \<sim> \<Gamma>"
  and "\<Gamma>2 \<sim> \<Gamma>"
  shows "\<Gamma>1 \<union> \<Gamma>2 \<sim> \<Gamma>"
using prems by (auto simp add: env_compat_def)

lemma env_compat_subset:
  assumes "\<Gamma>1 \<sim> \<Gamma>2"
  and "\<Gamma>1' \<subseteq> \<Gamma>1"
  shows "\<Gamma>1' \<sim> \<Gamma>2"
using prems by (auto simp add: env_compat_def)

lemma compat_ok_sanity:
  assumes "\<Gamma>1 \<union> \<Gamma>2 \<Turnstile> OK"
  and "\<Gamma>1 \<sim> \<Gamma>2"
  and "(x\<rhd>\<tau>) \<in> \<Gamma>1"
  and "qualT \<tau> = qL"
  shows "\<not> \<Gamma>2 defines x"
proof
  assume "\<Gamma>2 defines x"
  then obtain b where "(x, b) \<in> \<Gamma>2" by (auto simp add: dom_def)
  then obtain \<tau>' where "(x\<rhd>\<tau>') \<in> \<Gamma>2" by (cases b, auto)
  also hence "\<tau>' = \<tau>" using prems by (auto simp add: wf_env_def)
  finally have "(x\<rhd>\<tau>) \<in> \<Gamma>1 \<inter> \<Gamma>2" by auto
  hence "qualT \<tau> \<preceq> qU" using prems env_compat_def[of \<Gamma>1 \<Gamma>2] by fast
  thus False using prems by auto
qed

lemma compat_ok_sanity_right:
  assumes "\<Gamma>1 \<union> \<Gamma>2 \<Turnstile> OK"
  and "\<Gamma>1 \<sim> \<Gamma>2"
  and "(x\<rhd>\<tau>) \<in> \<Gamma>2"
  and "qualT \<tau> = qL"
  shows "\<not> \<Gamma>1 defines x"
using prems by (auto simp add: env_compat_symm[of \<Gamma>1 \<Gamma>2] Un_commute[of \<Gamma>1 \<Gamma>2]
                compat_ok_sanity[of \<Gamma>2 \<Gamma>1])

consts Typing :: "(env * env * tm * ty) set"
syntax
  "_Typing" :: "env \<Rightarrow> env \<Rightarrow> tm \<Rightarrow> ty \<Rightarrow> bool"   ("_;_ \<turnstile> _ : _" [50,50,80,80] 50)
translations
  "\<Gamma>;M \<turnstile> e : \<tau>" \<rightleftharpoons> "(\<Gamma>, M, e, \<tau>) \<in> Typing"

inductive Typing
  intros
  t_unit[intro!]  : "\<lbrakk> \<Gamma> \<Turnstile> OK; qualG \<Gamma> \<preceq> qU; M \<Turnstile> OK \<rbrakk> \<Longrightarrow>
                       \<Gamma>;M \<turnstile> TmUnit q : Qual q TyUnit"
  t_var[intro!]   : "\<lbrakk> \<Gamma> \<Turnstile> OK; (x\<rhd>\<tau>) \<in> \<Gamma>; qualG (\<Gamma>\<bar>x) \<preceq> qU; M \<Turnstile> OK  \<rbrakk> \<Longrightarrow>
                       \<Gamma>;M \<turnstile> TmFreeVar x : \<tau>"
  t_abs[intro!]   : "\<lbrakk> \<Gamma> \<Turnstile> OK; finite L; qualG \<Gamma> \<preceq> q; M \<Turnstile> OK;
                       \<forall>x. \<not> x free in e \<and> x \<notin> L \<and> \<not> \<Gamma> defines x \<longrightarrow>
                         \<Gamma>,(x:\<tau>1);M \<turnstile> freshen_tm x e : \<tau>2 \<rbrakk> \<Longrightarrow>
                       \<Gamma>;M \<turnstile> \<lambda>^q[\<tau>1]. e : Qual q (\<tau>1\<rightarrow>\<tau>2)"
  t_app[intro!]   : "\<lbrakk> \<Gamma>1 \<sim> \<Gamma>2; \<Gamma>1 \<union> \<Gamma>2 \<Turnstile> OK;
                       \<Gamma>1;M \<turnstile> e1 : Qual q (\<tau>1\<rightarrow>\<tau>2); \<Gamma>2;M \<turnstile> e2 : \<tau>1 \<rbrakk> \<Longrightarrow>
                       \<Gamma>1\<union>\<Gamma>2;M \<turnstile> e1\<cdot>e2 : \<tau>2"
  t_new[intro!]   : "\<lbrakk> \<Gamma>;M \<turnstile> e : \<tau> \<rbrakk> \<Longrightarrow>
                       \<Gamma>;M \<turnstile> TmNew e : Qual qL (TyRef \<tau>)"
  t_free[intro!]  : "\<lbrakk> \<Gamma>;M \<turnstile> e : Qual qL (TyRef \<tau>) \<rbrakk> \<Longrightarrow>
                       \<Gamma>;M \<turnstile> TmFree e : \<tau>"
  t_assign[intro!]: "\<lbrakk> \<Gamma>1 \<sim> \<Gamma>2; \<Gamma>1 \<union> \<Gamma>2 \<Turnstile> OK; qualT \<tau> \<preceq> qU;
                       \<Gamma>1;M \<turnstile> e1 : Qual qU (TyRef \<tau>); \<Gamma>2;M \<turnstile> e2 : \<tau> \<rbrakk> \<Longrightarrow>
                       \<Gamma>1\<union>\<Gamma>2;M \<turnstile> e1 := e2 : Qual qU TyUnit"
  t_swap[intro!]:   "\<lbrakk> \<Gamma>1 \<sim> \<Gamma>2; \<Gamma>1 \<union> \<Gamma>2 \<Turnstile> OK; qL \<preceq> qualT \<tau>;
                       \<Gamma>1;M \<turnstile> e1 : Qual qU (TyRef \<tau>); \<Gamma>2;M \<turnstile> e2 : \<tau> \<rbrakk> \<Longrightarrow>
                       \<Gamma>1\<union>\<Gamma>2;M \<turnstile> e1 :=: e2 : \<tau>"
  t_deref[intro!] : "\<lbrakk> \<Gamma>;M \<turnstile> e : Qual qU (TyRef \<tau>); qualT \<tau> \<preceq> qU \<rbrakk> \<Longrightarrow>
                       \<Gamma>;M \<turnstile> TmDeref e : \<tau>"
  t_loc[intro!]   : "\<lbrakk> \<Gamma> \<Turnstile> OK; qualG \<Gamma> \<preceq> qU; M \<Turnstile> OK; (l\<rhd>\<tau>) \<in> M \<rbrakk> \<Longrightarrow>
                       \<Gamma>;M \<turnstile> TmLoc q l : Qual q (TyRef \<tau>)"

inductive_cases typing_unit_elim [elim]:   "\<Gamma>;M \<turnstile> TmUnit q : \<tau>"
inductive_cases typing_var_elim [elim]:    "\<Gamma>;M \<turnstile> TmFreeVar x : \<tau>"
inductive_cases typing_abs_elim [elim]:    "\<Gamma>;M \<turnstile> \<lambda>^q[\<tau>1]. e : \<tau>2"
inductive_cases typing_app_elim [elim]:    "\<Gamma>;M \<turnstile> e1\<cdot>e2 : \<tau>"
inductive_cases typing_new_elim [elim]:    "\<Gamma>;M \<turnstile> TmNew e : \<tau>"
inductive_cases typing_free_elim [elim]:   "\<Gamma>;M \<turnstile> TmFree e : \<tau>"
inductive_cases typing_assign_elim [elim]: "\<Gamma>;M \<turnstile> e1 := e2 : \<tau>"
inductive_cases typing_swap_elim [elim]:   "\<Gamma>;M \<turnstile> e1 :=: e2 : \<tau>"
inductive_cases typing_deref_elim [elim]:  "\<Gamma>;M \<turnstile> TmDeref e : \<tau>"
inductive_cases typing_loc_elim [elim]:    "\<Gamma>;M \<turnstile> TmLoc q l : \<tau>"

lemma ty_wf_env:
  assumes "\<Gamma>;M \<turnstile> e : \<tau>"
  shows "\<Gamma> \<Turnstile> OK"
using prems by (induct, auto)

lemma ty_wf_env_loc:
  assumes "\<Gamma>;M \<turnstile> e : \<tau>"
  shows "M \<Turnstile> OK"
using prems by (induct, auto)

lemma ty_mem_strengthen:
  assumes "\<Gamma>;M \<turnstile> e : \<tau>"
  and "\<not> l free loc in e"
  shows "\<Gamma>;M\<bar>l \<turnstile> e : \<tau>"
proof -
  have "M \<Turnstile> OK" using prems ty_wf_env_loc by auto
  hence "M\<bar>l \<Turnstile> OK" using wf_env_subset[of M "M\<bar>l"] by (auto simp add: without_def)
  from prems this show ?thesis proof induct
    case (t_abs L M \<Gamma> \<tau>1 \<tau>2 e q)
    thus ?case proof auto
      fix x
      assume H1: "\<not> x free in e" and H2: "x \<notin> L" and H3: "\<not> \<Gamma> defines x"
      hence "\<not> l free loc in (vsubst_tm (TmFreeVar x) 0 e)" using t_abs vsubst_tm_same_loc by auto
      from t_abs.hyps H1 H2 H3 this t_abs.prems
      show "\<Gamma>,(x:\<tau>1);M\<bar>l \<turnstile> vsubst_tm (TmFreeVar x) 0 e : \<tau>2" by auto
    qed
  qed (auto simp add: without_def)
qed

lemma ty_mem_strengthen_all:
  assumes "\<Gamma>;M \<turnstile> e : \<tau>"
  and "\<forall>l. \<not> l free loc in e"
  shows "\<Gamma>;{} \<turnstile> e : \<tau>"
proof -
  have Hmok: "M \<Turnstile> OK" using prems ty_wf_env_loc by auto
  hence "finite M" by (auto simp add: wf_env_def)
  thus ?thesis using Hmok prems proof induct
    case (insert p M)
    then obtain l b where "p = (l, b)" by (cases p)
    then obtain \<tau>' where Hp: "p = (l\<rhd>\<tau>')" by (cases b, auto)
    have H: "M \<Turnstile> OK" using insert.prems wf_env_subset[of "insert p M" M] by auto
    have "\<Gamma>;(insert p M)\<bar>l \<turnstile> e : \<tau>" using prems by (auto simp add: ty_mem_strengthen)
    also have "(insert p M)\<bar>l = M\<bar>l" using Hp by (auto simp add: without_def)
    also have "\<not> M defines l" proof
      assume "M defines l"
      hence "p \<in> M" using Hp insert.prems wf_env_subset_defines[of "insert p M" "M" l \<tau>'] by auto
      thus False by contradiction
    qed
    hence "M\<bar>l = M" by (auto simp add: dom_def without_def)
    finally show ?case using insert H by auto
  qed auto
qed

lemma ty_no_loc:
  assumes "\<Gamma>; M \<turnstile> e : \<tau>"
  and "\<not> M defines l"
  shows "\<not> l free loc in e"
using prems proof induct
  case (t_abs L M \<Gamma> \<tau>1 \<tau>2 e q)
  hence "finite (dom \<Gamma>)" by (auto simp add: ty_wf_env wf_env_def dom_finite)
  then obtain x where "\<not> x free in e \<and> x \<notin> L \<and> \<not> \<Gamma> defines x"
    using fresh_exists[of "set (free_tm e) \<union> L \<union> dom \<Gamma>"] t_abs by auto
  hence "\<not> l free loc in (freshen_tm x e)" using t_abs by auto
  hence "\<not> l free loc in e" using vsubst_tm_same_loc by auto
  thus ?case by auto
qed (auto simp add: dom_def)

lemma ty_no_locs:
  assumes "\<Gamma>; {} \<turnstile> e : \<tau>"
  shows "\<forall>l. \<not> l free loc in e"
using prems by (auto simp add: ty_no_loc dom_def)

lemma free_ty_in_env:
  assumes "\<Gamma>;M \<turnstile> e : \<tau>"
  and "x free in e"
  shows "\<Gamma> defines x"
using prems proof induct
  case (t_abs L M \<Gamma> \<tau>1 \<tau>2 eb q)
  then obtain z where H: "z \<notin> set (free_tm eb) \<union> L \<union> dom \<Gamma> \<union> {x}"
    using fresh_exists[of "set (free_tm eb) \<union> L \<union> dom \<Gamma> \<union> {x}"]
    by (auto simp add: wf_env_def dom_finite)
  have "x free in freshen_tm z eb" using t_abs by (auto simp add: free_subst)
  thus ?case using t_abs.hyps H by (auto simp add: dom_def)
qed (auto simp add: dom_def)

lemma not_in_env_ty_same_subst:
  assumes "\<not> \<Gamma> defines x"
  and "\<Gamma>;M \<turnstile> vsubst_tm (TmFreeVar x) i e : \<tau>"
  shows "\<Gamma>;M \<turnstile> vsubst_tm e_s i e : \<tau>"
proof -
  have "\<not> x free in vsubst_tm (TmFreeVar x) i e" using prems by (auto simp add: free_ty_in_env)
  hence "\<forall>e_s. vsubst_tm e_s i e = e" using not_free_subst_same_subst by fast
  thus ?thesis using prems by auto
qed

lemma wf_env_defines_without:
  assumes "\<Gamma> \<Turnstile> OK"
  and "(x\<rhd>\<tau>) \<in> \<Gamma>"
  shows "\<Gamma> = (\<Gamma>\<bar>x),(x:\<tau>)"
using prems proof (auto simp add: without_def)
  fix y b
  assume "(y\<rhd>\<tau>) \<in> \<Gamma>" and "(y, b) \<in> \<Gamma>" and "b \<noteq> Term \<tau>"
  then obtain \<tau>' where "b = Term \<tau>'" by (cases b)
  hence "\<tau> = \<tau>'" using prems by (auto simp add: wf_env_def)
  thus False using prems by auto
qed

end
